#pragma once

#include <iostream>
#include <vector>
#include <set>


typedef struct Node 
{
	int nodeID = 0;
	std::vector<double> cords;
	bool vertex = 0;

	Node() = default;
	Node(int id, double x_, double y_, double z_, bool vertex_) :
		nodeID(id), cords{ x_,y_,z_ }, vertex(vertex_) {}

	//Node(int ID_, double x_, double y_, double z_, bool vertex_) :
		//nodeID(ID_), x(x_), y(y_), z(z_), vertex(vertex_) {}

	friend std::ostream& operator<<(std::ostream&, const Node&);
	//friend std::ostream& operator<<(std::ostream&, const std::set<Node>&);
	friend std::ostream& operator<<(std::ostream&, const std::vector<Node>&);

	bool operator<(const Node& p_node) const 
	{
		return nodeID < p_node.nodeID;
	}
	
	bool operator==(const Node& p_node) const 
	{
		if (nodeID == p_node.nodeID && cords == p_node.cords && vertex == p_node.vertex)
			return true;
		else return false;
	};
}Node;

typedef struct FiniteElement 
{
	int elementID = 0;
	int materialID = 0;
	mutable std::vector<int> nodesID;// ,    set

	friend std::ostream& operator<<(std::ostream&, const FiniteElement&);
	//friend std::ostream& operator<<(std::ostream&, const std::set<FiniteElement>&);
	friend std::ostream& operator<<(std::ostream&, const std::vector<FiniteElement>&);

	bool operator<(const FiniteElement& p_node) const 
	{
		return elementID < p_node.elementID;
	}

}FiniteElement;

typedef struct BoundaryFiniteElement : public FiniteElement // boundaryFEs (1 )
{//    
	int boundaryID = 0;// 
	int surfaceID = 0;//id  

	friend std::ostream& operator<<(std::ostream&, const BoundaryFiniteElement&);
	//,k,,.,,friend std::ostream& operator<<(std::ostream&, const std::multiset<BoundaryFiniteElement>&);
	friend std::ostream& operator<<(std::ostream&, const std::vector<BoundaryFiniteElement>&);

	bool operator==(const BoundaryFiniteElement& p_node) const
	{
		return boundaryID == p_node.boundaryID;
	}

	bool operator<(const BoundaryFiniteElement& p_node) const
	{
		return boundaryID < p_node.boundaryID;
	}
}BoundaryFiniteElement;

typedef struct Edge //    
{ 
	int first_node_id;
	int last_node_id;
	int center_id;

	Edge() = default;
	Edge(int id1, int id2, int id3) :
		first_node_id(id1), last_node_id(id2), center_id(id3) {}

	friend inline bool operator==(const Edge& left_edge, const Edge& right_edge) 
	{
		if (right_edge.first_node_id == left_edge.first_node_id && right_edge.last_node_id == left_edge.last_node_id)
			return true;
		else if (right_edge.first_node_id == left_edge.last_node_id && right_edge.last_node_id == left_edge.first_node_id)
			return true;
		else return false;
	}

	bool operator<(const Edge& right_edge) const 
	{
		return center_id < right_edge.center_id;
	}
	void updateCenter(int id_) 
	{
		center_id = id_;
	}
}Edge;



template < typename _T >
inline void hashCombine(std::size_t& o_seed, const _T& p_val)
{
	o_seed ^= std::hash <_T >() (p_val) + 0x9e3779b9 + (o_seed << 6) + (o_seed >> 2);
}
template < typename _T >
inline void hashVal(std::size_t& o_seed, const _T& p_val)
{
	hashCombine(o_seed, p_val);
}

namespace std {
	template<>
	class hash<Edge> {
	public:
		size_t operator()(const Edge& s) const
		{

			int left(s.first_node_id);
			int right(s.last_node_id);
			size_t seed = 0;

			if (left > right) {
				swap(left, right);
			}

			hashCombine(seed, left);
			hashCombine(seed, right);

			return seed;
		}
	};
}
