#include <iostream>
#include <string>
#include <fstream>
#include <algorithm>
#include <memory>
#include "NeuMeshLoader.h"
#include "MeshLoader.h"
#include "AneuMeshLoader.h"

//!!!    ( !)
//!!!    std::ranges 
//!!! 1.        .
//!!! 2.    .
//!!! 3.         ID  
//!!!       (..   10    10).



MeshLoader* File_Format_Checker(const std::string& cfile)
{ 
	std::string file = cfile;
    while (file.back() == ' ')                     
        file.pop_back();
    MeshLoader* mesh_pointer = nullptr;

    // check .aneu
    std::size_t found = file.rfind(".aneu");
    if (found != std::string::npos && file.length() - found == 5) 
	{
        std::cout << "File-format-checker was passed." << std::endl;
        //dynamic_cast<AneuMeshLoader*>(mesh_pointer);
        mesh_pointer = new AneuMeshLoader;
        return mesh_pointer;
    }

    // check .neu
    found = file.rfind(".neu");
    if (found != std::string::npos && file.length() - found == 4) 
	{
        std::cout << "File-format-checker was passed." << std::endl;
        //dynamic_cast<NeuMeshLoader*>(mesh_pointer);   
        mesh_pointer = new NeuMeshLoader;//smart 
        return mesh_pointer;
    }

    else 
	{
        std::cout << "Warning! File-format-checker was not passed!" << std::endl;
        return 0;
        //std::size_t dot = file.rfind('.');
        /*if (dot == std::string::npos)
            throw BadFileFormatException();
        else throw BadFileFormatException(file.substr(file.rfind('.'), file.size() - file.rfind('.')));*/
    };
}

void Mesh_Loader_Demo(MeshLoader* loader_, const std::string& filepath_) 
{
	//AneuMeshLoader mesh;
    loader_->LoadMesh(filepath_);
    loader_->Print_Data();
    /*
    //std::vector<FiniteElement> elements_vector;
    std::vector<int> elements_vector;
    std::set<int> nodes_vector;
    std::vector<std::set<int>> neighbours;
    std::vector<BoundaryFiniteElement> boundaryFEs_vector;

    std::cout << "Result of getElementsByID(2, 4, 8) function:\n";
    elements_vector = loader_->getElementsByID(2, 4, 8);
    std::cout << elements_vector;
    elements_vector.clear();

    std::cout << "\nResult of getElementsByEdge(2, 8) function:\n";
    elements_vector = loader_->getElementsByEdge(2, 8);
    std::cout << elements_vector;
    elements_vector.clear();

    std::cout << "\nResult of getNodesByBoundary(3) function:\n";
    nodes_vector = loader_->getNodesByBoundary(3);
    std::cout << nodes_vector;
    nodes_vector.clear();

    std::cout << "\nResult of getelementsByBoundary(4) function:\n";
    elements_vector = loader_->getElementsByBoundary(4);
    std::cout << elements_vector;
    elements_vector.clear();

    std::cout << "\nResult of getTrianglesAndNodes(1):\n";
    boundaryFEs_vector = loader_->getTrianglesAndNodes(1);
    std::cout << boundaryFEs_vector;
    boundaryFEs_vector.clear();

    std::cout << "\nResult of getsboundaryFEsByBoundary (1) function:\n";
    boundaryFEs_vector = loader_->getsboundaryFEsByBoundary(1);
    std::cout << boundaryFEs_vector;
    boundaryFEs_vector.clear();
	std::cout << "\nResult of getNodeNeighbours() function:\n";
	neighbours = loader_->getNodeNeighbours();
	for (auto i = 0; i < neighbours.size(); ++i) {
		std::cout << "Neighbours for Node with ID : " << i + 1 << std::endl;
		for (const auto& j : neighbours[i]) {
			std::cout << j << std::endl;
		}
	}
	neighbours.clear();


	std::cout << "\nResult of insertCenterNode() function:\n";
	loader_->insertCenterNode();
    loader_->Print_Data();
    */

    std::vector<Node> nodes = loader_->getNodes();
    std::vector<FiniteElement> finite_elements = loader_->getFiniteElements();
    std::vector<BoundaryFiniteElement> boundaries = loader_->getBoundaryFiniteElements();


    std::vector<int> elems_by_ID = loader_->getElementsByID(1, 2, 3);
    std::vector<int> elems_by_edge = loader_->getElementsByEdge(1, 2);

    std::set<int> nodes_by_surface = loader_->getNodesByBoundary(1);

    std::set<int> boundary_by_surface = loader_->getElementsByBoundary(2);

    loader_->insertCenterNode();

    auto neighbor_nodes = loader_->getNodeNeighbours();

}

/*struct fixed_tolower : std::unary_function<char, char> //   
{ 
    char operator()(char c) const 
	{
        return tolower((unsigned char)c);
    }
};*/


int main(int argc, char* argv[]) 
{
    std::string filepath = "cube.aneu";
    if (argc == 1)
        std::cout << "No arguments were added, using standard file PATH..." << std::endl;
    else if (argc > 2)
        std::cout << "There must be one or none arguments! Set file path to default..." << std::endl;
    else 
	{
        std::cout << "Set file path to \"" << argv[1] << "\"..." << std::endl;
        filepath = argv[1];
        //std::transform(filepath.begin(), filepath.end(), filepath.begin(), fixed_tolower()); // .aneu = .ANEU = .AnEu = etc
    };

    MeshLoader* loader = File_Format_Checker(filepath);
    Mesh_Loader_Demo(loader, filepath);
	delete loader;
    return 0;
}