﻿#include <algorithm>
#include <stdlib.h>
#include <unordered_set>
#include <iterator>
#include <map>
#include "MeshLoader.h"
#include <fstream>
#include <ranges>

const std::vector<Node>& MeshLoader::getNodes() 
{
	return nodes;
}

const std::vector<FiniteElement>& MeshLoader::getFiniteElements() 
{
	return elements;
}

const std::vector<BoundaryFiniteElement>& MeshLoader::getBoundaryFiniteElements() 
{
	return boundaryFEs;
}


//!!! Контенеры-результаты должны содержать только ID узлов или КЭ


std::vector<int> MeshLoader::getElementsByID(const int id1, const int id2, const int id3) //возможность нескольких узлов - контейнер
{
	std::vector<int> result;
	auto current_iter = elements.begin();
	while (current_iter != elements.end())
	{
		current_iter = std::find_if(current_iter, elements.end(), [id1, id2, id3](const FiniteElement& current_element) {
			return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end()
				&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id3) != current_element.nodesID.end(); });
		
		if (current_iter != elements.end())
		{
			result.push_back(current_iter->elementID);
			current_iter++;
		}
	}
	return result;//O(n)
}

std::vector<int> MeshLoader::getElementsByEdge(const int id1, const int id2) //возможность нескольких КЭ 
{
	/*std::vector<FiniteElement> result;
	std::copy_if(elements.begin(), elements.end(), std::back_inserter(result), [id1, id2](const FiniteElement& current_element) {
		return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end()
			&& std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id2) != current_element.nodesID.end(); });
	return result;*/
	std::vector<int> result;
	auto   last = boundaryFEs.begin();
	while (last != boundaryFEs.end())
	{
		last = find_if(last, boundaryFEs.end(), [id1, id2](const FiniteElement& elem)
			{
				return (std::find(elem.nodesID.begin(), elem.nodesID.end(), id1) != elem.nodesID.end()
					&& std::find(elem.nodesID.begin(), elem.nodesID.end(), id2) != elem.nodesID.end());
			});

		if (last != boundaryFEs.end())
			result.push_back(last->elementID);
		last++;
	}

	return result;

}
/*
std::vector<int> MeshLoader::getElementsByOneID(const int id1) 
{
	std::vector<int> result;
	std::copy_if(elements.begin(), elements.end(), std::back_inserter(result), [id1](const FiniteElement& current_element) {
		return std::find(current_element.nodesID.begin(), current_element.nodesID.end(), id1) != current_element.nodesID.end(); });
	return result;
	//O(n)
}*/

std::vector<int> MeshLoader::getElementsByOneID(const int id1) 
{
	std::vector<int> result;

	auto it_beg = elements.begin();
	auto it_end = elements.end();

	while (it_beg != it_end) {
		auto it_res = std::find_if(it_beg, it_end, [id1](const FiniteElement& f_el) {
			return f_el.elementID == id1;
			});
		it_beg = it_res;
		if (it_res != it_end) {
			result.push_back(it_res->elementID); 
			++it_beg;
		}
	}

	return result;
}


std::set<int> MeshLoader::getNodesByBoundary(int boundaryID) 
{
	std::set<int> result;

	auto first = boundaryFEs.begin();
	auto last = boundaryFEs.end();

	while (first != last) {
		auto it_res = std::find_if(first, last, [boundaryID](const BoundaryFiniteElement& b_el) {
			return b_el.boundaryID == boundaryID;
			});
		first = it_res;
		if (it_res != last) {
			for (const auto& b_node : it_res->nodesID) 
			{
				result.insert(b_node);
			}
			++first;
		}
	}

	return result;
}

std::set<int> MeshLoader::getElementsByBoundary(int id) 
{
	std::set<int> result;

	auto first = boundaryFEs.begin();
	auto last = boundaryFEs.end();

	while (first != last) 
	{
		auto it_res = std::find_if(first, last, [id](const BoundaryFiniteElement& b_el) {
			return b_el.boundaryID == id;
			});
		first = it_res;
		if (it_res != last) 
		{
			result.insert(it_res->elementID);
			++first;
		}
	}
	return result;
}

std::vector<int> MeshLoader::getsboundaryFEsByBoundary(int boundaryID)
{
	std::vector<int> result;
	auto last = boundaryFEs.begin();
	while (last != boundaryFEs.end()) {
		last = std::find_if(last, boundaryFEs.end(), [boundaryID](const BoundaryFiniteElement& elem) {
			return elem.boundaryID == boundaryID; });
		if (last != boundaryFEs.end()) {
			result.push_back(last->elementID);
			++last;
		}
	}
	return result;

}

void MeshLoader::insertCenterNode() {
	
	std::unordered_set<Edge> edges;

	for (auto& it : elements) {
		std::vector<int> current_nodes(it.nodesID);
		for (auto i = current_nodes.begin(); i < current_nodes.end() - 1; ++i)
			for (auto j = i + 1; j < current_nodes.end(); ++j) {
				Edge current_edge(*i, *j, 0);

				auto founded_edge = edges.find(current_edge);

				if (founded_edge == edges.cend()) {
					Node left_node = nodes[*i - 1];
					Node right_node = nodes[*j - 1];
					int center_id_ = nodes.rbegin()->nodeID + 1;
					Node center_node(center_id_, (left_node.cords[0] + right_node.cords[0]) / 2,
						(left_node.cords[1] + right_node.cords[1]) / 2, (left_node.cords[2] + right_node.cords[2]) / 2, false);

					current_edge.updateCenter(center_id_);
					nodes.push_back(center_node);
					edges.insert(current_edge);
					it.nodesID.push_back(center_id_);
				}
				else {
					it.nodesID.push_back(founded_edge->center_id);
				}
			}
		it.nodesID.shrink_to_fit();
	}



	for (auto& it : boundaryFEs) {
		std::vector<int> current_nodes(it.nodesID);
		for (auto i = current_nodes.begin(); i < current_nodes.end() - 1; ++i)
			for (auto j = i + 1; j < current_nodes.end(); ++j) {
				Edge current_edge(*i, *j, 0);

				auto founded_edge = edges.find(current_edge);

				if (founded_edge != edges.cend())
					it.nodesID.push_back(founded_edge->center_id);
			}
		it.nodesID.shrink_to_fit();
	}

	boundaryFEs.shrink_to_fit();
	nodes.shrink_to_fit();

}


std::vector<std::set<int>> MeshLoader::getNodeNeighbours()
{
	std::vector<std::set<int>> result(getNodes().size());

	for (auto& i : elements) {
		for (int j : i.nodesID) {
			result[j - 1].insert(i.nodesID.begin(), i.nodesID.end());
			for (std::set<int>::iterator iter = result[j - 1].begin(); iter != result[j - 1].end();)
			{
				if (*iter == j)
				{
					iter = result[j - 1].erase(iter);
				}
				else
				{
					++iter;
				}
			}
		}
	}
	return result;

}


void MeshLoader::Print_Data() 
{
	std::cout << getNodes() << std::endl
		<< getFiniteElements() << std::endl
		<< getBoundaryFiniteElements() << std::endl;
}


const std::vector<double>& MeshLoader::findNodeRanges(int id)
{
	return (std::ranges::find_if(nodes.begin(), nodes.end(), [id](const Node& elem)
		{
			return elem.nodeID == id;
		}))->cords;
}


std::vector<Node>::iterator MeshLoader::getNodeByID(int id)
{
	Node tmp;
	tmp.nodeID = id;
	return find(nodes.begin(), nodes.end(), tmp);
}



std::vector<BoundaryFiniteElement> MeshLoader::getTrianglesAndNodes(int surfaceID)
{
	std::vector<BoundaryFiniteElement> result;//vector?
	std::copy_if(boundaryFEs.begin(), boundaryFEs.end(), std::back_inserter(result), 
	[surfaceID](const BoundaryFiniteElement& current_element) {
		return surfaceID == current_element.surfaceID;
		});
	auto square = [](const Node& A, const Node& B, const Node& C)
	{
		
		double a = sqrt((C.cords[0] - B.cords[0]) * (C.cords[0] - B.cords[0])
			+ (C.cords[1] - B.cords[1]) * (C.cords[1] - B.cords[1]) 
			+ (C.cords[2] - B.cords[2]) * (C.cords[2] - B.cords[2]));
		double b = sqrt((C.cords[0] - A.cords[0]) * (C.cords[0] - A.cords[0])
			+ (C.cords[1] - A.cords[1]) * (C.cords[1] - A.cords[1]) 
			+ (C.cords[2] - A.cords[2]) * (C.cords[2] - A.cords[2]));
		double c = sqrt((B.cords[0] - A.cords[0]) * (B.cords[0] - A.cords[0]) 
			+ (B.cords[1] - A.cords[1]) * (B.cords[1] - A.cords[1]) 
			+ (B.cords[2] - A.cords[2]) * (B.cords[2] - A.cords[2]));
		double p = 1 / 2 * (a + b + c);
		return sqrt(p * (p - a) * (p - b) * (p - c));
	};
	
	std::ranges::sort(result, [&](const BoundaryFiniteElement& a, const BoundaryFiniteElement& b)
		{
			Node X1 = *getNodeByID(a.nodesID[0]);
			Node Y1 = *getNodeByID(a.nodesID[1]);
			Node Z1 = *getNodeByID(a.nodesID[2]);

			Node X2 = *getNodeByID(b.nodesID[0]);
			Node Y2 = *getNodeByID(b.nodesID[1]);
			Node Z2 = *getNodeByID(b.nodesID[2]);

			return square(X1,Y1,Z1) > square(X2,Y2,Z2); 
		});
	std::ofstream triangle("triangle.txt", std::ofstream::out | std::ofstream::trunc);

	std::ostream_iterator<BoundaryFiniteElement> os(triangle, " ");
	
	//!!! В задании на защиту подразумевалось использовать std::ranges::filter для вывода узлов.
	std::ranges::copy(result | std::ranges::views::take(10) | std::views::drop(10), std::ostream_iterator<BoundaryFiniteElement>(triangle, "\n"));
	
	triangle.close();
	return result;//O(n)
}
namespace std
{
	std::ostream& operator<< (std::ostream& p_out, const std::vector<int>& vec) {
		for (size_t i : std::ranges::views::iota(size_t(0), vec.size()))
			p_out << vec[i] << " ";

		p_out << std::endl;
		return p_out;
	}
}